@@ -44,6 +44,8 @@ rpl_slave_state *rpl_global_gtid_slave_state;
 /* Object used for MASTER_GTID_WAIT(). */
 gtid_waiting rpl_global_gtid_waiting;
 
+const char *const Relay_log_info::state_delaying_string = "Waiting until MASTER_DELAY seconds after master executed event";
+
 Relay_log_info::Relay_log_info(bool is_slave_recovery, const char* thread_name)
   :Slave_reporting_capability(thread_name),
    replicate_same_server_id(::replicate_same_server_id),
@@ -1009,8 +1011,7 @@ void Relay_log_info::inc_group_relay_log_pos(ulonglong log_pos,
     {
       if (cmp < 0)
       {
-        safe_strcpy(group_master_log_name, sizeof(group_master_log_name),
-                    rgi->future_event_master_log_name);
+        strcpy(group_master_log_name, rgi->future_event_master_log_name);
         group_master_log_pos= log_pos;
       }
       else if (group_master_log_pos < log_pos)
@@ -1631,7 +1632,8 @@ scan_one_gtid_slave_pos_table(THD *thd, HASH *hash, DYNAMIC_ARRAY *array,
       goto end;
     }
 
-    if ((rec= my_hash_search(hash, (const uchar *)&domain_id, 0)))
+    if ((rec= my_hash_search(hash, (const uchar *)&domain_id,
+                             sizeof(domain_id))))
     {
       entry= (struct gtid_pos_element *)rec;
       if (entry->sub_id >= sub_id)
@@ -1931,7 +1933,7 @@ rpl_load_gtid_slave_state(THD *thd)
   for (i= 0; i < array.elements; ++i)
   {
     get_dynamic(&array, (uchar *)&tmp_entry, i);
-    if ((err= rpl_global_gtid_slave_state->update_nolock(tmp_entry.gtid.domain_id,
+    if ((err= rpl_global_gtid_slave_state->update(tmp_entry.gtid.domain_id,
                                                   tmp_entry.gtid.server_id,
                                                   tmp_entry.sub_id,
                                                   tmp_entry.gtid.seq_no,
@@ -2252,7 +2254,7 @@ delete_or_keep_event_post_apply(rpl_group_info *rgi,
 }
 
 
-void rpl_group_info::cleanup_context(THD *thd, bool error, bool keep_domain_owner)
+void rpl_group_info::cleanup_context(THD *thd, bool error)
 {
   DBUG_ENTER("rpl_group_info::cleanup_context");
   DBUG_PRINT("enter", ("error: %d", (int) error));
@@ -2286,9 +2288,11 @@ void rpl_group_info::cleanup_context(THD *thd, bool error, bool keep_domain_owne
 
   if (unlikely(error))
   {
-    // leave alone any XA prepared transactions
-    if (thd->transaction->xid_state.is_explicit_XA() &&
-        thd->transaction->xid_state.get_state_code() != XA_PREPARED)
+    /*
+      trans_rollback above does not rollback XA transactions
+      (todo/fixme consider to do so.
+    */
+    if (thd->transaction->xid_state.is_explicit_XA())
       xa_trans_force_rollback(thd);
 
     thd->release_transactional_locks();
@@ -2307,7 +2311,7 @@ void rpl_group_info::cleanup_context(THD *thd, bool error, bool keep_domain_owne
       Ensure we always release the domain for others to process, when using
       --gtid-ignore-duplicates.
     */
-    if (gtid_ignore_duplicate_state != GTID_DUPLICATE_NULL && !keep_domain_owner)
+    if (gtid_ignore_duplicate_state != GTID_DUPLICATE_NULL)
       rpl_global_gtid_slave_state->release_domain_owner(this);
   }
 
@@ -2386,17 +2390,7 @@ void rpl_group_info::slave_close_thread_tables(THD *thd)
 {
   DBUG_ENTER("rpl_group_info::slave_close_thread_tables(THD *thd)");
   thd->get_stmt_da()->set_overwrite_status(true);
-#ifdef WITH_WSREP
-  // This can happen e.g. when table_def::compatible_with fails and sets a error
-  // but thd->is_error() is false then. However, we do not want to commit
-  // statement on Galera instead we want to rollback it as later in
-  // apply_write_set we rollback transaction and that can't be done
-  // after wsrep transaction state is s_committed.
-  if (WSREP(thd))
-    (thd->is_error() || thd->is_slave_error) ? trans_rollback_stmt(thd) : trans_commit_stmt(thd);
-  else
-#endif
-    thd->is_error() ? trans_rollback_stmt(thd) : trans_commit_stmt(thd);
+  thd->is_error() ? trans_rollback_stmt(thd) : trans_commit_stmt(thd);
   thd->get_stmt_da()->set_overwrite_status(false);
 
   close_thread_tables(thd);
@@ -2436,13 +2430,8 @@ mark_start_commit_inner(rpl_parallel_entry *e, group_commit_orderer *gco,
   uint64 count= ++e->count_committing_event_groups;
   /* Signal any following GCO whose wait_count has been reached now. */
   tmp= gco;
-
-  DBUG_ASSERT(!tmp->gc_done);
-
   while ((tmp= tmp->next_gco))
   {
-    DBUG_ASSERT(!tmp->gc_done);
-
     uint64 wait_count= tmp->wait_count;
     if (wait_count > count)
       break;
