@@ -107,39 +107,23 @@ public:
 class Wsrep_load_data_split
 {
 public:
-  Wsrep_load_data_split(THD *thd, TABLE *table)
+  Wsrep_load_data_split(THD *thd)
     : m_thd(thd)
-    , m_load_data_splitting(false)
+    , m_load_data_splitting(wsrep_load_data_splitting)
     , m_fragment_unit(thd->wsrep_trx().streaming_context().fragment_unit())
     , m_fragment_size(thd->wsrep_trx().streaming_context().fragment_size())
   {
-    /*
-      We support load data splitting for InnoDB only as it will use
-      streaming replication (SR).
-    */
-    if (WSREP(thd) && wsrep_load_data_splitting)
+    if (WSREP(m_thd) && m_load_data_splitting)
     {
-      // For partitioned tables find underlying hton
-      handlerton *ht= table->file->partition_ht();
-      if (ht->db_type != DB_TYPE_INNODB)
-      {
-        push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
-                            ER_NOT_SUPPORTED_YET,
-                            "wsrep_load_data_splitting for other than InnoDB tables");
-      }
-      else
-      {
-        /* Override streaming settings with backward compatible values for
-           load data splitting */
-        m_thd->wsrep_cs().streaming_params(wsrep::streaming_context::row, 10000);
-        m_load_data_splitting= true;
-      }
+      /* Override streaming settings with backward compatible values for
+         load data splitting */
+      m_thd->wsrep_cs().streaming_params(wsrep::streaming_context::row, 10000);
     }
   }
 
   ~Wsrep_load_data_split()
   {
-    if (m_load_data_splitting)
+    if (WSREP(m_thd) && m_load_data_splitting)
     {
       /* Restore original settings */
       m_thd->wsrep_cs().streaming_params(m_fragment_unit, m_fragment_size);
@@ -271,10 +255,6 @@ public:
   */
   void skip_data_till_eof()
   {
-#ifndef EMBEDDED_LIBRARY
-    if (mysql_bin_log.is_open())
-      cache.read_function= cache.real_read_function;
-#endif
     while (GET != my_b_EOF)
       ;
   }
@@ -364,7 +344,6 @@ int mysql_load(THD *thd, const sql_exchange *ex, TABLE_LIST *table_list,
   bool is_concurrent;
 #endif
   const char *db= table_list->db.str;		// This is never null
-
   /*
     If path for file is not defined, we will use the current database.
     If this is not set, we will use the directory where the table to be
@@ -375,6 +354,9 @@ int mysql_load(THD *thd, const sql_exchange *ex, TABLE_LIST *table_list,
   bool transactional_table __attribute__((unused));
   DBUG_ENTER("mysql_load");
 
+#ifdef WITH_WSREP
+  Wsrep_load_data_split wsrep_load_data_split(thd);
+#endif /* WITH_WSREP */
   /*
     Bug #34283
     mysqlbinlog leaves tmpfile after termination if binlog contains
@@ -439,11 +421,6 @@ int mysql_load(THD *thd, const sql_exchange *ex, TABLE_LIST *table_list,
   {
     DBUG_RETURN(TRUE);
   }
-
-#ifdef WITH_WSREP
-  Wsrep_load_data_split wsrep_load_data_split(thd, table_list->table);
-#endif /* WITH_WSREP */
-
   thd_proc_info(thd, "Executing");
   /*
     Let us emit an error if we are loading data to table which is used
@@ -677,12 +654,12 @@ int mysql_load(THD *thd, const sql_exchange *ex, TABLE_LIST *table_list,
         (!table->triggers ||
          !table->triggers->has_delete_triggers()))
         table->file->extra(HA_EXTRA_WRITE_CAN_REPLACE);
-    if (thd->locked_tables_mode <= LTM_LOCK_TABLES &&
-        !table->s->long_unique_table)
+    if (thd->locked_tables_mode <= LTM_LOCK_TABLES)
       table->file->ha_start_bulk_insert((ha_rows) 0);
     table->copy_blobs=1;
 
     thd->abort_on_warning= !ignore && thd->is_strict_mode();
+    thd->get_stmt_da()->reset_current_row_for_warning(1);
 
     bool create_lookup_handler= handle_duplicates != DUP_ERROR;
     if ((table_list->table->file->ha_table_flags() & HA_DUPLICATE_POS))
