@@ -97,15 +97,18 @@ char*	srv_data_home;
 char*	srv_undo_dir;
 
 /** The number of tablespaces to use for rollback segments. */
-ulong	srv_undo_tablespaces;
+uint	srv_undo_tablespaces;
 
 /** The number of UNDO tablespaces that are open and ready to use. */
-ulint	srv_undo_tablespaces_open;
+uint32_t srv_undo_tablespaces_open;
 
 /** The number of UNDO tablespaces that are active (hosting some rollback
 segment). It is quite possible that some of the tablespaces doesn't host
 any of the rollback-segment based on configuration used. */
-ulint	srv_undo_tablespaces_active;
+uint32_t srv_undo_tablespaces_active;
+
+/** Rate at which UNDO records should be purged. */
+ulong	srv_purge_rseg_truncate_frequency;
 
 /** Enable or Disable Truncate of UNDO tablespace.
 Note: If enabled then UNDO tablespace will be selected for truncate.
@@ -159,7 +162,7 @@ uint		srv_flush_log_at_timeout;
 /** innodb_page_size */
 ulong		srv_page_size;
 /** log2 of innodb_page_size; @see innodb_init_params() */
-ulong		srv_page_size_shift;
+uint32_t	srv_page_size_shift;
 /** innodb_log_write_ahead_size */
 ulong		srv_log_write_ahead_size;
 
@@ -395,6 +398,9 @@ mysql_mutex_t srv_misc_tmpfile_mutex;
 /** Temporary file for miscellanous diagnostic output */
 FILE*	srv_misc_tmpfile;
 
+static ulint	srv_main_thread_process_no;
+static ulint	srv_main_thread_id;
+
 /* The following counts are used by the srv_master_callback. */
 
 /** Iterations of the loop bounded by 'srv_active' label. */
@@ -412,6 +418,17 @@ thread ensures that we flush the log files at least once per
 second. */
 static time_t	srv_last_log_flush_time;
 
+/* Interval in seconds at which various tasks are performed by the
+master thread when server is active. In order to balance the workload,
+we should try to keep intervals such that they are not multiple of
+each other. For example, if we have intervals for various tasks
+defined as 5, 10, 15, 60 then all tasks will be performed when
+current_time % 60 == 0 and no tasks will be performed when
+current_time % 5 != 0. */
+
+# define	SRV_MASTER_CHECKPOINT_INTERVAL		(7)
+# define	SRV_MASTER_DICT_LRU_INTERVAL		(47)
+
 /** Buffer pool dump status frequence in percentages */
 ulong srv_buf_dump_status_frequency;
 
@@ -503,8 +520,26 @@ static srv_sys_t	srv_sys;
 struct purge_coordinator_state
 {
   /** Snapshot of the last history length before the purge call.*/
-  size_t history_size;
+  size_t m_history_length;
   Atomic_counter<int> m_running;
+private:
+  ulint count;
+  ulint n_use_threads;
+  ulint n_threads;
+
+  ulint lsn_lwm;
+  ulint lsn_hwm;
+  ulonglong start_time;
+  ulint lsn_age_factor;
+
+  static constexpr ulint adaptive_purge_threshold= 20;
+  static constexpr ulint safety_net= 20;
+  ulint series[innodb_purge_threads_MAX + 1];
+
+  inline void compute_series();
+  inline void lazy_init();
+  void refresh(bool full);
+
 public:
   inline void do_purge();
 };
@@ -654,7 +689,6 @@ void srv_boot()
   if (transactional_lock_enabled())
     sql_print_information("InnoDB: Using transactional memory");
 #endif
-  buf_dblwr.init();
   srv_thread_pool_init();
   trx_pool_init();
   srv_init();
@@ -861,7 +895,12 @@ srv_printf_innodb_monitor(
 			n_reserved);
 	}
 
-	fprintf(file, "Process ID=0, Main thread ID=0, state: %s\n",
+	fprintf(file,
+		"Process ID=" ULINTPF
+		", Main thread ID=" ULINTPF
+		", state: %s\n",
+		srv_main_thread_process_no,
+		srv_main_thread_id,
 		srv_main_thread_op_info);
 	fprintf(file,
 		"Number of rows inserted " ULINTPF
@@ -981,14 +1020,17 @@ srv_export_innodb_status(void)
 
 	export_vars.innodb_data_writes = os_n_file_writes;
 
-	buf_dblwr.lock();
-	ulint dblwr = buf_dblwr.written();
-	export_vars.innodb_dblwr_pages_written = dblwr;
-	export_vars.innodb_dblwr_writes = buf_dblwr.batches();
-	buf_dblwr.unlock();
+	ulint dblwr = 0;
 
-	export_vars.innodb_data_written = srv_stats.data_written
-		+ (dblwr << srv_page_size_shift);
+	if (buf_dblwr.is_initialised()) {
+		buf_dblwr.lock();
+		dblwr = buf_dblwr.submitted();
+		export_vars.innodb_dblwr_pages_written = buf_dblwr.written();
+		export_vars.innodb_dblwr_writes = buf_dblwr.batches();
+		buf_dblwr.unlock();
+	}
+
+	export_vars.innodb_data_written = srv_stats.data_written + dblwr;
 
 	export_vars.innodb_buffer_pool_read_requests
 		= buf_pool.stat.n_page_gets;
@@ -996,11 +1038,41 @@ srv_export_innodb_status(void)
 	export_vars.innodb_buffer_pool_write_requests =
 		srv_stats.buf_pool_write_requests;
 
+	export_vars.innodb_buffer_pool_reads = srv_stats.buf_pool_reads;
+
+	export_vars.innodb_buffer_pool_read_ahead_rnd =
+		buf_pool.stat.n_ra_pages_read_rnd;
+
+	export_vars.innodb_buffer_pool_read_ahead =
+		buf_pool.stat.n_ra_pages_read;
+
+	export_vars.innodb_buffer_pool_read_ahead_evicted =
+		buf_pool.stat.n_ra_pages_evicted;
+
+	export_vars.innodb_buffer_pool_pages_data =
+		UT_LIST_GET_LEN(buf_pool.LRU);
+
 	export_vars.innodb_buffer_pool_bytes_data =
 		buf_pool.stat.LRU_bytes
 		+ (UT_LIST_GET_LEN(buf_pool.unzip_LRU)
 		   << srv_page_size_shift);
 
+	export_vars.innodb_buffer_pool_pages_dirty =
+		UT_LIST_GET_LEN(buf_pool.flush_list);
+
+	export_vars.innodb_buffer_pool_pages_made_young
+		= buf_pool.stat.n_pages_made_young;
+	export_vars.innodb_buffer_pool_pages_made_not_young
+		= buf_pool.stat.n_pages_not_made_young;
+
+	export_vars.innodb_buffer_pool_pages_old = buf_pool.LRU_old_len;
+
+	export_vars.innodb_buffer_pool_bytes_dirty =
+		buf_pool.stat.flush_list_bytes;
+
+	export_vars.innodb_buffer_pool_pages_free =
+		UT_LIST_GET_LEN(buf_pool.free);
+
 #ifdef UNIV_DEBUG
 	export_vars.innodb_buffer_pool_pages_latched =
 		buf_get_latched_pages_number();
@@ -1036,9 +1108,10 @@ srv_export_innodb_status(void)
 
 	export_vars.innodb_row_lock_current_waits= lock_sys.get_wait_pending();
 
-	export_vars.innodb_row_lock_time = lock_sys.get_wait_time_cumulative();
-	export_vars.innodb_row_lock_time_max = lock_sys.get_wait_time_max();
-
+	export_vars.innodb_row_lock_time = lock_sys.get_wait_time_cumulative()
+		/ 1000;
+	export_vars.innodb_row_lock_time_max = lock_sys.get_wait_time_max()
+		/ 1000;
 	mysql_mutex_unlock(&lock_sys.wait_mutex);
 
 	export_vars.innodb_row_lock_time_avg= export_vars.innodb_row_lock_waits
@@ -1263,13 +1336,6 @@ bool srv_any_background_activity()
 
 static void purge_worker_callback(void*);
 static void purge_coordinator_callback(void*);
-static void purge_truncation_callback(void*)
-{
-  purge_sys.latch.rd_lock(SRW_LOCK_CALL);
-  const purge_sys_t::iterator head= purge_sys.head;
-  purge_sys.latch.rd_unlock();
-  head.free_history();
-}
 
 static tpool::task_group purge_task_group;
 tpool::waitable_task purge_worker_task(purge_worker_callback, nullptr,
@@ -1277,48 +1343,60 @@ tpool::waitable_task purge_worker_task(purge_worker_callback, nullptr,
 static tpool::task_group purge_coordinator_task_group(1);
 static tpool::waitable_task purge_coordinator_task
   (purge_coordinator_callback, nullptr, &purge_coordinator_task_group);
-static tpool::task_group purge_truncation_task_group(1);
-static tpool::waitable_task purge_truncation_task
-  (purge_truncation_callback, nullptr, &purge_truncation_task_group);
+
+static tpool::timer *purge_coordinator_timer;
 
 /** Wake up the purge threads if there is work to do. */
-void purge_sys_t::wake_if_not_active()
+void
+srv_wake_purge_thread_if_not_active()
 {
-  if (enabled() && !paused() && !purge_state.m_running &&
-      (srv_undo_log_truncate || trx_sys.history_exists()) &&
-      ++purge_state.m_running == 1)
-    srv_thread_pool->submit_task(&purge_coordinator_task);
+	ut_ad(!srv_read_only_mode);
+
+	if (purge_sys.enabled() && !purge_sys.paused()
+	    && trx_sys.history_exists()) {
+		if(++purge_state.m_running == 1) {
+			srv_thread_pool->submit_task(&purge_coordinator_task);
+		}
+	}
 }
 
 /** @return whether the purge tasks are active */
-bool purge_sys_t::running()
+bool purge_sys_t::running() const
 {
   return purge_coordinator_task.is_running();
 }
 
-void purge_sys_t::stop_FTS()
+/** Suspend purge in data dictionary tables */
+void purge_sys_t::stop_SYS()
 {
   latch.rd_lock(SRW_LOCK_CALL);
-  m_FTS_paused++;
+  ++m_SYS_paused;
   latch.rd_unlock();
-  while (m_active)
-    std::this_thread::sleep_for(std::chrono::seconds(1));
 }
 
 /** Stop purge during FLUSH TABLES FOR EXPORT */
 void purge_sys_t::stop()
 {
-  latch.wr_lock(SRW_LOCK_CALL);
-
-  if (!enabled())
+  for (;;)
   {
-    /* Shutdown must have been initiated during FLUSH TABLES FOR EXPORT. */
-    ut_ad(!srv_undo_sources);
-    latch.wr_unlock();
-    return;
-  }
+    latch.wr_lock(SRW_LOCK_CALL);
 
-  ut_ad(srv_n_purge_threads > 0);
+    if (!enabled())
+    {
+      /* Shutdown must have been initiated during FLUSH TABLES FOR EXPORT. */
+      ut_ad(!srv_undo_sources);
+      latch.wr_unlock();
+      return;
+    }
+
+    ut_ad(srv_n_purge_threads > 0);
+
+    if (!must_wait_SYS())
+      break;
+
+    latch.wr_unlock();
+    std::this_thread::sleep_for(std::chrono::seconds(1));
+  }
 
   const auto paused= m_paused++;
 
@@ -1335,8 +1413,9 @@ void purge_sys_t::stop()
 /** Resume purge in data dictionary tables */
 void purge_sys_t::resume_SYS(void *)
 {
-  ut_d(auto paused=) purge_sys.m_SYS_paused--;
-  ut_ad(paused);
+  ut_d(const auto s=)
+  purge_sys.m_SYS_paused--;
+  ut_ad(s);
 }
 
 /** Resume purge at UNLOCK TABLES after FLUSH TABLES FOR EXPORT */
@@ -1358,8 +1437,8 @@ void purge_sys_t::resume()
    if (paused == 1)
    {
      ib::info() << "Resuming purge";
-     purge_state.m_running= 1;
-     srv_thread_pool->submit_task(&purge_coordinator_task);
+     purge_state.m_running = 0;
+     srv_wake_purge_thread_if_not_active();
      MONITOR_ATOMIC_INC(MONITOR_PURGE_RESUME_COUNT);
    }
    latch.wr_unlock();
@@ -1408,29 +1487,31 @@ static void srv_sync_log_buffer_in_background()
 	}
 }
 
-/** Report progress during shutdown.
-@param last   time of last output
-@param n_read number of page reads initiated for change buffer merge */
-static void srv_shutdown_print(time_t &last, ulint n_read)
+/*********************************************************************//**
+This function prints progress message every 60 seconds during server
+shutdown, for any activities that master thread is pending on. */
+static
+void
+srv_shutdown_print_master_pending(
+/*==============================*/
+	time_t*		last_print_time,	/*!< last time the function
+						print the message */
+	ulint		n_bytes_merged)		/*!< number of change buffer
+						just merged */
 {
-  time_t now= time(nullptr);
-  if (now - last >= 15)
-  {
-    last= now;
+	time_t current_time = time(NULL);
 
-    const ulint ibuf_size= ibuf.size;
-    sql_print_information("Completing change buffer merge;"
-                          " %zu page reads initiated;"
-                          " %zu change buffer pages remain",
-                          n_read, ibuf_size);
-#if defined HAVE_SYSTEMD && !defined EMBEDDED_LIBRARY
-    service_manager_extend_timeout(INNODB_EXTEND_TIMEOUT_INTERVAL,
-                                   "Completing change buffer merge;"
-                                   " %zu page reads initiated;"
-                                   " %zu change buffer pages remain",
-                                   n_read, ibuf_size);
-#endif
-  }
+	if (difftime(current_time, *last_print_time) > 60) {
+		*last_print_time = current_time;
+
+		/* Check change buffer merge, we only wait for change buffer
+		merge if it is a slow shutdown */
+		if (!srv_fast_shutdown && n_bytes_merged) {
+			ib::info() << "Waiting for change buffer merge to"
+				" complete number of bytes of change buffer"
+				" just merged: " << n_bytes_merged;
+		}
+	}
 }
 
 /** Perform periodic tasks whenever the server is active.
@@ -1441,7 +1522,7 @@ static void srv_master_do_active_tasks(ulonglong counter_time)
 
 	MONITOR_INC(MONITOR_MASTER_ACTIVE_LOOPS);
 
-	if (!(counter_time % (47 * 1000000ULL))) {
+	if (!(counter_time % (SRV_MASTER_DICT_LRU_INTERVAL * 1000000ULL))) {
 		srv_main_thread_op_info = "enforcing dict cache limit";
 		if (ulint n_evicted = dict_sys.evict_table_LRU(true)) {
 			MONITOR_INC_VALUE(
@@ -1475,7 +1556,7 @@ Complete the shutdown tasks such as background DROP TABLE,
 and optionally change buffer merge (on innodb_fast_shutdown=0). */
 void srv_shutdown(bool ibuf_merge)
 {
-	ulint		n_read = 0;
+	ulint		n_bytes_merged	= 0;
 	time_t		now = time(NULL);
 
 	do {
@@ -1484,16 +1565,21 @@ void srv_shutdown(bool ibuf_merge)
 		++srv_main_shutdown_loops;
 
 		if (ibuf_merge) {
-			srv_main_thread_op_info = "doing insert buffer merge";
-			/* Disallow the use of change buffer to
-			avoid a race condition with
-			ibuf_read_merge_pages() */
-			ibuf_max_size_update(0);
+			srv_main_thread_op_info = "checking free log space";
 			log_free_check();
-			n_read = ibuf_contract();
-			srv_shutdown_print(now, n_read);
+			srv_main_thread_op_info = "doing insert buffer merge";
+			n_bytes_merged = ibuf_merge_all();
+
+			/* Flush logs if needed */
+			srv_sync_log_buffer_in_background();
 		}
-	} while (n_read);
+
+		/* Print progress message every 60 seconds during shutdown */
+		if (srv_print_verbose_log) {
+			srv_shutdown_print_master_pending(&now,
+							  n_bytes_merged);
+		}
+	} while (n_bytes_merged);
 }
 
 /** The periodic master task controlling the server. */
@@ -1504,7 +1590,8 @@ void srv_master_callback(void*)
   ut_a(srv_shutdown_state <= SRV_SHUTDOWN_INITIATED);
 
   MONITOR_INC(MONITOR_MASTER_THREAD_SLEEP);
-  purge_sys.wake_if_not_active();
+  if (!purge_state.m_running)
+    srv_wake_purge_thread_if_not_active();
   ulonglong counter_time= microsecond_interval_timer();
   srv_sync_log_buffer_in_background();
   MONITOR_INC_TIME_IN_MICRO_SECS(MONITOR_SRV_LOG_FLUSH_MICROSECOND,
@@ -1519,7 +1606,7 @@ void srv_master_callback(void*)
 }
 
 /** @return whether purge should exit due to shutdown */
-static bool srv_purge_should_exit(size_t old_history_size)
+static bool srv_purge_should_exit()
 {
   ut_ad(srv_shutdown_state <= SRV_SHUTDOWN_CLEANUP);
 
@@ -1530,12 +1617,8 @@ static bool srv_purge_should_exit(size_t old_history_size)
     return true;
 
   /* Slow shutdown was requested. */
-  size_t prepared, active= trx_sys.any_active_transactions(&prepared);
   const size_t history_size= trx_sys.history_size();
-
-  if (!history_size);
-  else if (!active && history_size == old_history_size && prepared);
-  else
+  if (history_size)
   {
     static time_t progress_time;
     time_t now= time(NULL);
@@ -1552,7 +1635,7 @@ static bool srv_purge_should_exit(size_t old_history_size)
     return false;
   }
 
-  return !active;
+  return !trx_sys.any_active_transactions();
 }
 
 /*********************************************************************//**
@@ -1581,80 +1664,190 @@ static bool srv_task_execute()
 
 static void purge_create_background_thds(int );
 
-/** Flag which is set, whenever innodb_purge_threads changes. */
-static Atomic_relaxed<bool> srv_purge_thread_count_changed;
-
-static std::mutex purge_thread_count_mtx;
+std::mutex purge_thread_count_mtx;
 void srv_update_purge_thread_count(uint n)
 {
 	std::lock_guard<std::mutex> lk(purge_thread_count_mtx);
-	purge_create_background_thds(n);
+	ut_ad(n > 0);
+	ut_ad(n <= innodb_purge_threads_MAX);
 	srv_n_purge_threads = n;
-	srv_purge_thread_count_changed = true;
+	srv_purge_thread_count_changed = 1;
 }
 
+Atomic_counter<int> srv_purge_thread_count_changed;
+
 inline void purge_coordinator_state::do_purge()
 {
   ut_ad(!srv_read_only_mode);
+  lazy_init();
+  ut_ad(n_threads);
+  bool wakeup= false;
 
-  if (!purge_sys.enabled() || purge_sys.paused())
-    return;
-
-  uint n_threads;
+  purge_coordinator_timer->disarm();
 
+  while (purge_sys.enabled() && !purge_sys.paused())
   {
-    std::lock_guard<std::mutex> lk(purge_thread_count_mtx);
-    n_threads= srv_n_purge_threads;
-    srv_purge_thread_count_changed= false;
-    goto first_loop;
-  }
+loop:
+    wakeup= false;
+    const auto now= my_interval_timer();
+    const auto sigcount= m_running;
+
+    if (now - start_time >= 1000000)
+    {
+      refresh(false);
+      start_time= now;
+    }
+
+    const auto old_activity_count= srv_sys.activity_count;
+    const auto history_size= trx_sys.history_size();
 
-  do
-  {
     if (UNIV_UNLIKELY(srv_purge_thread_count_changed))
     {
       /* Read the fresh value of srv_n_purge_threads, reset
-      the changed flag. Both are protected by purge_thread_count_mtx. */
+      the changed flag. Both are protected by purge_thread_count_mtx.
+
+      This code does not run concurrently, it is executed
+      by a single purge_coordinator thread, and no races
+      involving srv_purge_thread_count_changed are possible. */
       {
         std::lock_guard<std::mutex> lk(purge_thread_count_mtx);
-        n_threads= srv_n_purge_threads;
-        srv_purge_thread_count_changed= false;
+        n_threads= n_use_threads= srv_n_purge_threads;
+        srv_purge_thread_count_changed= 0;
+      }
+      refresh(true);
+      start_time= now;
+    }
+    else if (history_size > m_history_length)
+    {
+      /* dynamically adjust the purge thread based on redo log fill factor */
+      if (n_use_threads < n_threads && lsn_age_factor < lsn_lwm)
+      {
+more_threads:
+        ++n_use_threads;
+        lsn_hwm= lsn_lwm;
+        lsn_lwm-= series[n_use_threads];
+      }
+      else if (n_use_threads > 1 && lsn_age_factor >= lsn_hwm)
+      {
+fewer_threads:
+        --n_use_threads;
+        lsn_lwm= lsn_hwm;
+        lsn_hwm+= series[n_use_threads];
+      }
+      else if (n_use_threads == 1 && lsn_age_factor >= 100 - safety_net)
+      {
+        wakeup= true;
+        break;
       }
     }
-  first_loop:
-    ut_ad(n_threads);
+    else if (n_threads > n_use_threads &&
+             srv_max_purge_lag && m_history_length > srv_max_purge_lag)
+      goto more_threads;
+    else if (n_use_threads > 1 && old_activity_count == srv_sys.activity_count)
+      goto fewer_threads;
 
-    history_size= trx_sys.history_size();
+    ut_ad(n_use_threads);
+    ut_ad(n_use_threads <= n_threads);
 
-    if (!history_size)
+    m_history_length= history_size;
+
+    if (history_size &&
+        trx_purge(n_use_threads,
+                  !(++count % srv_purge_rseg_truncate_frequency) ||
+                  purge_sys.truncate.current ||
+                  (srv_shutdown_state != SRV_SHUTDOWN_NONE &&
+                   srv_fast_shutdown == 0)))
+      continue;
+
+    if (m_running == sigcount)
     {
-    no_history:
-      srv_dml_needed_delay= 0;
-      purge_truncation_task.wait();
-      trx_purge_truncate_history();
+      /* Purge was not woken up by srv_wake_purge_thread_if_not_active() */
+
+      /* The magic number 5000 is an approximation for the case where we have
+      cached undo log records which prevent truncate of rollback segments. */
+      wakeup= history_size &&
+        (history_size >= 5000 ||
+         history_size != trx_sys.history_size_approx());
       break;
     }
-
-    ulint n_pages_handled= trx_purge(n_threads, history_size);
-    if (!trx_sys.history_exists())
-      goto no_history;
-    if (purge_sys.truncating_tablespace() ||
-        srv_shutdown_state != SRV_SHUTDOWN_NONE)
-    {
-      purge_truncation_task.wait();
-      trx_purge_truncate_history();
-    }
-    else
-      srv_thread_pool->submit_task(&purge_truncation_task);
-    if (!n_pages_handled)
+    else if (!trx_sys.history_exists())
       break;
+
+    if (!srv_purge_should_exit())
+      goto loop;
   }
-  while (purge_sys.enabled() && !purge_sys.paused() &&
-         !srv_purge_should_exit(history_size));
+
+  if (wakeup)
+    purge_coordinator_timer->set_time(10, 0);
 
   m_running= 0;
 }
 
+inline void purge_coordinator_state::compute_series()
+{
+  ulint points= n_threads;
+  memset(series, 0, sizeof series);
+  constexpr ulint spread= 100 - adaptive_purge_threshold - safety_net;
+
+  /* We distribute spread across n_threads,
+  e.g.: spread of 60 is distributed across n_threads=4 as: 6+12+18+24 */
+
+  const ulint additional_points= (points * (points + 1)) / 2;
+  if (spread % additional_points == 0)
+  {
+    /* Arithmetic progression is possible. */
+    const ulint delta= spread / additional_points;
+    ulint growth= delta;
+    do
+    {
+      series[points--]= growth;
+      growth += delta;
+    }
+    while (points);
+    return;
+  }
+
+  /* Use average distribution to spread across the points */
+  const ulint delta= spread / points;
+  ulint total= 0;
+  do
+  {
+    series[points--]= delta;
+    total+= delta;
+  }
+  while (points);
+
+  for (points= 1; points <= n_threads && total++ < spread; )
+    series[points++]++;
+}
+
+inline void purge_coordinator_state::lazy_init()
+{
+  if (n_threads)
+    return;
+  n_threads= n_use_threads= srv_n_purge_threads;
+  refresh(true);
+  start_time= my_interval_timer();
+}
+
+void purge_coordinator_state::refresh(bool full)
+{
+  if (full)
+  {
+    compute_series();
+    lsn_lwm= adaptive_purge_threshold;
+    lsn_hwm= adaptive_purge_threshold + series[n_threads];
+  }
+
+  mysql_mutex_lock(&log_sys.mutex);
+  const lsn_t last= log_sys.last_checkpoint_lsn,
+    max_age= log_sys.max_checkpoint_age;
+  mysql_mutex_unlock(&log_sys.mutex);
+
+  lsn_age_factor= ulint(((log_sys.get_lsn() - last) * 100) / max_age);
+}
+
+
 static std::list<THD*> purge_thds;
 static std::mutex purge_thd_mutex;
 extern void* thd_attach_thd(THD*);
@@ -1719,13 +1912,16 @@ static void purge_coordinator_callback(void*)
 
 void srv_init_purge_tasks()
 {
-  purge_create_background_thds(srv_n_purge_threads);
-  purge_sys.coordinator_startup();
+  purge_create_background_thds(innodb_purge_threads_MAX);
+  purge_coordinator_timer= srv_thread_pool->create_timer
+    (purge_coordinator_callback, nullptr);
 }
 
 static void srv_shutdown_purge_tasks()
 {
   purge_coordinator_task.disable();
+  delete purge_coordinator_timer;
+  purge_coordinator_timer= nullptr;
   purge_worker_task.wait();
   std::unique_lock<std::mutex> lk(purge_thd_mutex);
   while (!purge_thds.empty())
@@ -1734,7 +1930,6 @@ static void srv_shutdown_purge_tasks()
     purge_thds.pop_front();
   }
   n_purge_thds= 0;
-  purge_truncation_task.wait();
 }
 
 /**********************************************************************//**
@@ -1774,22 +1969,15 @@ ulint srv_get_task_queue_length()
 /** Shut down the purge threads. */
 void srv_purge_shutdown()
 {
-  if (purge_sys.enabled())
-  {
-    if (!srv_fast_shutdown && !opt_bootstrap)
-    {
-      srv_purge_batch_size= innodb_purge_batch_size_MAX;
-      srv_update_purge_thread_count(innodb_purge_threads_MAX);
-    }
-    size_t history_size= trx_sys.history_size();
-    while (!srv_purge_should_exit(history_size))
-    {
-      history_size= trx_sys.history_size();
-      ut_a(!purge_sys.paused());
-      srv_thread_pool->submit_task(&purge_coordinator_task);
-      purge_coordinator_task.wait();
-    }
-    purge_sys.coordinator_shutdown();
-    srv_shutdown_purge_tasks();
-  }
+	if (purge_sys.enabled()) {
+		if (!srv_fast_shutdown && !opt_bootstrap)
+			srv_update_purge_thread_count(innodb_purge_threads_MAX);
+		while(!srv_purge_should_exit()) {
+			ut_a(!purge_sys.paused());
+			srv_wake_purge_thread_if_not_active();
+			purge_coordinator_task.wait();
+		}
+		purge_sys.coordinator_shutdown();
+		srv_shutdown_purge_tasks();
+	}
 }
